from typing import Dict
from Vertex import Vertex
import collections
from utils import print_clique
import math
import random
import TempGraph
import time


class StaticGraph:

    # This class represents a static graph (intersection graph) of a temporal graph for a time interval [a,b]
    def __init__(self, name="unnamed"):
        # contains a dictionary of all vertices
        self.vertexList = collections.OrderedDict()  # type: collections.OrderedDict[int, Vertex]
        # contains number of vertices
        self.numVertices = 0
        # contains dictionary mapping vertices to adjacency lists
        self.edgeList = {}  # type: Dict[Vertex, list[Vertex]]
        # contains numbe rof edges
        self.numEdges = 0
        # give the graph a specific name
        self.name = name
        # contains a mapping from the original names in the input file to the names vertex names in the program
        self.name_space = []
        # represents the start time a of the interval
        self.start_time = 0
        # represents the end time b of the interval
        self.end_time = 0

    # set the time interval [a,b]
    def set_timeframe(self, a, b):
        self.start_time = a
        self.end_time = b

    # add a vertex to the graph
    def addVertex(self, key):
        if self.__contains__(key):
            return None
        self.numVertices += 1
        vertex = Vertex(key)
        self.vertexList[key] = vertex
        self.edgeList[vertex] = []
        return vertex

    # remove a vertex from the graph
    def removeVertex(self, key):
        if self.__contains__(key):
            return None
        self.numVertices -= 1
        vertex = self.vertexList.pop(key)
        self.edgeList[vertex].clear()
        for neighbor in self.getNeighbors(vertex):
            self.edgeList[neighbor].remove(vertex)

    # get the vertex corresponding to the id
    def getVertex(self, id):
        if id in self.vertexList:
            return self.vertexList[id]
        else:
            return None


    def __contains__(self, vertex):
        return vertex in self.vertexList

    # add an edge to the static graph, mostly similiar to the temporal version
    def addEdge(self, u, v):
        # type: (int, int, int) -> None
        if u not in self.vertexList:
            self.addVertex(u)
            self.edgeList[self.vertexList[u]] = []
        if v not in self.vertexList:
            self.addVertex(v)
            self.edgeList[self.vertexList[v]] = []
        u = self.vertexList[u]
        v = self.vertexList[v]
        if v not in self.edgeList[u] and u not in self.edgeList[v]:
            self.edgeList[u].append(v)
            self.edgeList[v].append(u)
            self.numEdges += 1

    # get the degree of a specific vertex
    def getDegree(self, vertex):
        # type: (Vertex) -> int
        return len(self.edgeList[vertex])

    # print a vertex with degree
    def printVertices(self):
        for vertex in self.vertexList.values():  # type: Vertex
            print("Vertex: " + str(vertex.name) + " Degree: " + str(self.getDegree(vertex)))

    # get all vertices
    def getVertices(self):
        return self.vertexList.values()

    # print all edges in an easy-to-read format
    def printEdges(self):
        for vertex in self.vertexList.values():
            for neighbor in self.edgeList[vertex]:
                if vertex.id < neighbor.id:
                    print("(" + str(vertex.name) + " " + str(neighbor.name) + ")")

    # print the set of neighbors for vertex in an easy-to-read format
    def printNeighbors(self, vertex):
        # type: (Vertex) -> None
        print("{", end='')
        for neighbor in self.edgeList[vertex]:
            print("{" + str(vertex.name) + "," + str(neighbor.name) + "}", end='')
        print("}", end='\n')

    # get a copy of the neighborhod of our vertex
    def getNeighbors(self, vertex):
        # type: (Vertex) -> list[Vertex]
        return self.edgeList[vertex].copy()

    # add another timestamp to the intersection graph.
    # Used in the following way:
    # We create a intersection graph for timestamp a
    # While iterating over the intervals [a,a+1],[a,a+2],...,[a,b] we just remove all edges which aren't
    # existing in the next timestamp (a+1),(a+2),...(b)
    # Thus, we don't need to create a new intersection graph for each interval
    def add_another_timestamp(self, temporal_graph, b):

        # type: (TempGraph, int) -> None
        # adds the timestamp [b] from the temporal graph to the current static graph
        # for every edge:
            # if it exists for timestamp b, keep it
            # remove otherwise
        for vertex in self.vertexList.values():
            if temporal_graph.edgeList[vertex].__contains__(b):
                i = 0
                length = len(self.edgeList[vertex])
                while i < length:
                    neighbor = self.edgeList[vertex][i]
                    if not temporal_graph.edgeList[vertex][b].__contains__(neighbor):
                        self.edgeList[vertex].remove(neighbor)
                        self.edgeList[neighbor].remove(vertex)
                        self.numEdges -= 1
                        i -= 1
                        length -= 1
                    i += 1
            else:
                length = len(self.edgeList[vertex])
                while length != 0:
                    neighbor = self.edgeList[vertex][0]
                    self.edgeList[vertex].remove(neighbor)
                    self.edgeList[neighbor].remove(vertex)
                    self.numEdges -= 1
                    length -= 1
        self.end_time = b

    # sort the set of vertices by degree and change the id of the vertices accordingly.
    # This is needed for the pivot procedure of enumerating cliques
    # We do update the id of the vertex, but keep track of the name so we can reverse it to the original names
    def sort_by_degree(self):
        ordering = sorted([vertex.name for vertex in self.vertexList.values()],
                          key=lambda x: len(self.edgeList[self.vertexList[x]]))
        name_space = {}
        new_vertexList = collections.OrderedDict()
        pos = 0
        for current in ordering:
            new_vertexList[pos] = self.vertexList[current]
            new_vertexList[pos].id = pos
            name_space[pos] = new_vertexList[pos].name
            pos += 1
        self.vertexList = new_vertexList
        self.name_space = name_space

    # Return number of neighbors of vertex in clique
    def number_of_neighbors_in_set(self, v, clique):
        # type: (Vertex, Dict[int, Vertex]) -> int
        number = 0
        for neighbor in self.edgeList[v]:
            if neighbor.id in clique.keys():
                number += 1
        return number

    # Return number of edges of group which are outgoing
    def outgoing_edges(self, group):
        # type: (Dict[int,Vertex]) -> int
        n = 0
        for vertex in group.values():
            for neighbor in self.getNeighbors(vertex):
                if neighbor not in group.values():
                    n += 1
        return n

    def outgoing_edges_clique(self, group, group2):
        n = 0
        for vertex in group.values():
            for neighbor in self.getNeighbors(vertex):
                if neighbor not in group2.values():
                    n += 1
        return n


    def strict_positive_neighborhood(self,vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]
        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id > vertex.id:
                tmp[neighbor.id] = neighbor
        return tmp


    def inclusive_positive_neighborhood(self, vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]
        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id > vertex.id:
                tmp[neighbor.id] = neighbor
        tmp[vertex.id] = vertex
        return tmp


    def strict_negative_neighborhood(self,vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]

        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id < vertex.id:
                tmp[neighbor.id] = neighbor
        return tmp


    def inclusive_negative_neighborhood(self, vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]

        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id < vertex.id:
                tmp[neighbor.id] = neighbor
        tmp[vertex.id] = vertex
        return tmp


    def is_neighbor(self, vertex1, vertex2):
        # type: (Vertex, Vertex) -> bool
        return self.edgeList[vertex1].__contains__(vertex2)

    # Create a complement graph of our static intersection graph.
    # This is used on the vertex cover sub-procedure.
    def complement_graph(self, vertex_set, name):
        # type: (Dict[int, Vertex], int) -> StaticGraph

        # initalize vertices
        complement_graph = StaticGraph("Complement" + str(name))
        for vertex in vertex_set.values():
            if vertex.name != name:
                newVertex = Vertex(vertex.id)
                complement_graph.vertexList[newVertex.id] = newVertex
                complement_graph.edgeList[newVertex] = []

        # add edge if it is not an edge in the original graph
        for vertex in vertex_set.values():
            if vertex.name != name:
                for vertex2 in vertex_set.values():
                    if vertex2.name != name and vertex2.name > vertex.name:
                        if not self.is_neighbor(vertex, vertex2):
                            complement_graph.edgeList[complement_graph.vertexList[vertex.id]].append(
                                complement_graph.vertexList[vertex2.id])
                            complement_graph.edgeList[complement_graph.vertexList[vertex2.id]].append(
                                complement_graph.vertexList[vertex.id])

        # return graph
        complement_graph.numVertices = len(complement_graph.vertexList)
        return complement_graph

    # Write static graph to a file format to compare with an algorithm for enumerating cliques on static graphs
    def graph_to_OCaml_file(self, filename):
        f = open(filename, "w")
        for vertex in self.vertexList.values():
            for neighbor in self.edgeList[vertex]:
                if vertex.id < neighbor.id:
                    f.write(str(vertex.name) + " " + str(neighbor.name) + "\n")


    def __iter__(self):
        return iter(self.vertexList.values())


    def print_graph(self):
        print("StaticGraph (" + self.name + "):")
        print("Vertices:")
        print(self.printVertices())
        print("Edges:")
        print(self.printEdges())
        print("Timeframe:")
        print(str(self.start_time) + " " + str(self.end_time))
        print("Namespace: ")
        print(self.name_space)



    def create_random_graph(self, nodes, edges, noise):

        edges = min(edges, (nodes * (nodes - 1)) / 2)
        remaining_edges = edges
        for i in range(0, nodes):
            self.addVertex(i)

        while remaining_edges > 0:
            max_clique_size = int(1 / 2 * (math.sqrt(8 * edges + 1) + 1))
            clique_size = random.randint(2, max_clique_size)
            clique = random.sample(range(0, nodes), clique_size)
            for vertex in clique:
                for neighbor in clique:
                    if vertex != neighbor:
                        if not self.edgeList[self.vertexList[vertex]].__contains__(
                                self.vertexList[neighbor]):
                            remaining_edges -= 1
                            self.addEdge(vertex, neighbor)
        remaining_edges = noise
        while remaining_edges > 0:
            edge = random.sample(range(0, nodes), 2)
            vertex = edge[0]
            neighbor = edge[1]
            if not self.edgeList[self.vertexList[vertex]].__contains__(
                    self.vertexList[neighbor]):
                remaining_edges -= 1
                self.addEdge(vertex, neighbor)

    # method to write the static graph to a file in temporal graph format
    def graph_to_Temp_file(self, path):
        f = open(path, "w")
        for vertex in self.vertexList.values():
            for neighbor in self.edgeList[vertex]:
                if vertex.id < neighbor.id:
                    f.write("0 " + str(vertex.name) + " " + str(neighbor.name) + "\n")
































