from typing import Dict
from Vertex import Vertex
import StaticGraph
import collections

class TempGraph:

    # This class represents a temporal graph
    def __init__(self, filename):
        # contains a dictionary of all vertices
        self.vertexList = collections.OrderedDict()    # type: collections.OrderedDict[int, Vertex]
        # contains number of vertices
        self.numVertices = 0
        # contains dictionary of all vertices mapped to a dictionary with timestamps mapped to adjacency lists
        self.edgeList = {}      # type: Dict[Vertex, collections.OrderedDict[int, list[Vertex]]]
        # contains number of edges
        self.numEdges = 0
        # contains the latest timestamp value
        self.t_max = 0
        # contains the earliest timestamp value
        self.t_min = 0
        # contain the name of the file used for input
        self.filename = filename
        # contains a mapping from the original timestamps in the input file to the timestamps used in the program
        self.time_space = {}
        # contains a mapping from the original names in the input file to the names vertex names in the program
        self.name_space = {}
        # reverse mapping as name_space for constant time look up
        self.reverse_name_space = []
        # the mode [0-5] represents the used subset variation (look isolated_subsets.py for mapping)
        self.mode = 0
        # contains the delta value used to create this graph (look delta-graph in paper)
        self.delta = 0
        # contains the version used to compute vertex/time maximal cliques (0 - old version, 1 - new version)
        self.maximality_variant = 1

        # contains a threshold of minimum size of isolated cliques (delete all cliques of size at most this value)
        self.del_small_clique = 1

        # variables which are counted durint computation, potential data for output
        # number of candidate sets computed in enumeration stage
        self.num_candidate_sets = 0
        # number of cliques found before removing non time/vertex maximal ones
        self.num_clique_before_time_max = 0
        # number o f overall intervalls iterated
        self.num_timestamps_iter = 0


    # add a vertex to the graph
    def addVertex(self, key):

        if self.__contains__(key):
            return None
        self.numVertices += 1
        vertex = Vertex(key)
        self.vertexList[key] = vertex
        self.edgeList[self.vertexList[key]] = collections.OrderedDict()
        return vertex

    # get vertex from graph. return none if not existing
    def getVertex(self, id):
        if id in self.vertexList:
            return self.vertexList[id]
        else:
            return None

    def __contains__(self, n):
        return n in self.vertexList

    # add edge for given timestamp to graph
    def addEdge(self, u, v, t=0):
        # type: (int, int, int, int) -> None

        # If vertex u/v is not part of graph yet, add it.
        if u not in self.vertexList:
            self.addVertex(u)
            self.edgeList[self.vertexList[u]] = collections.OrderedDict()
        if v not in self.vertexList:
            self.addVertex(v)
            self.edgeList[self.vertexList[v]] = collections.OrderedDict()
        u = self.vertexList[u]
        v = self.vertexList[v]
        # initialize adjacency list for u/v at timestamp t if not existing yet
        if t not in self.edgeList[u]:
            self.edgeList[u][t] = []
        if t not in self.edgeList[v]:
            self.edgeList[v][t] = []

        # add edge if not part of graph yet
        if v not in self.edgeList[u][t] and u not in self.edgeList[v][t]:
            self.edgeList[u][t].append(v)
            self.edgeList[v][t].append(u)
            self.numEdges += 1

    # returns whether graph contains edge (u,v,t)
    def has_edge(self, t, u, v):
        if t in self.edgeList[self.vertexList[u]]:
            if self.vertexList[v] in self.edgeList[self.vertexList[u]][t]:
                return True
        return False

    # set the value of the maximum timestamp
    def set_t_max(self, t_max):
        self.t_max = t_max

    # set the value of the minimum timestamp
    def set_t_min(self, t_min):
        self.t_min = t_min

    # get the number of edges outgoing from given vertex for all timestamps combined
    def getDegree(self, vertex):
        # type: (Vertex) -> int
        num = 0
        for timestamps in self.edgeList[vertex].values():
            num += len(timestamps)
        return num

    # get degree but limited to a specific timestamp
    def getDegree_for_timestamp(self, vertex, timestamp):
        if self.edgeList[vertex].keys().__contains__(timestamp):
            return len(self.edgeList[vertex][timestamp])
        else:
            return 0

    # print (mostly for debug used) of all vertices with the respective degree
    def printVertices(self):
        for vertex in self.vertexList.values():  # type: Vertex
            print("Vertex: " + str(vertex.name) + " Degree: " + str(self.getDegree(vertex)))

    # get list of all vertices
    def getVertices(self):
        return self.vertexList.keys()

    # print all edges in an easy-to-read format
    def printEdges(self):
        for vertex in self.vertexList.values():
            for time in self.edgeList[vertex].keys():
                for neighbor in self.edgeList[vertex][time]:
                    if vertex.id < neighbor.id:
                        print("(" + str(time) + " " + str(vertex.name) + " " + str(neighbor.name) + ")")

    # print the list of neighbors of a given vertex in an easy-to-read format
    def printNeighbors(self, vertex):
        # type: (Vertex) -> None
        print("{", end='')
        for time in self.edgeList[vertex].keys():
            for neighbor in self.edgeList[vertex][time]:
                print("{" + str(vertex.id) + "," + str(neighbor.id) + "}", end='')
        print("}", end='\n')

    # return the list of neighbors for given vertex
    def getNeighbors(self, vertex):
        # type: (TempGraph, Vertex) -> list[Vertex]

        tmp = []
        for time in self.edgeList[vertex].keys():
            for neighbor in self.edgeList[vertex][time]:
                tmp.append(neighbor)
        return tmp


    # Return number of neighbors of vertex in clique
    def number_of_neighbors_in_set(self, v, clique):
        # type: (Vertex, collections.OrderedDict[int, Vertex]) -> int
        n = 0
        for timestamps in self.edgeList[v].values():
            for neighbor in timestamps:
                if neighbor.id in clique.keys():
                    n += 1
        return n


    # Return number of edges of group which are outgoing for given time interval
    def outgoing_edges(self, group, a, b):
        # type: (collections.OrderedDict[int,Vertex], int, int) -> int

        outgoing_edges = 0

        for vertex in group.values():
            for curr_time in (a, b+1):
                try:
                    for neighbor in self.edgeList[vertex][curr_time]:
                        if neighbor not in group.keys():
                            outgoing_edges += 1
                except KeyError:
                    continue

        return outgoing_edges

    # Return number of edges from vertex which are outgoing from group for given time interval
    def outgoing_edges_from_vertex(self, vertex, group, a, b):
        # type: (Vertex, collections.OrderedDict[int,Vertex], int, int) -> int

        outgoing_edges = 0

        for curr_time in (a, b+1):
            try:
                for neighbor in self.edgeList[vertex][curr_time]:
                    if neighbor not in group.keys():
                        outgoing_edges += 1
            except KeyError:
                continue
        return outgoing_edges


    # compute the static graph (intersection graph) for interval [a,b]
    def getStaticGraph(self, a, b):
        # type: (TempGraph, int, int) -> StaticGraph

        # initialize static graph
        lifetime = b - a + 1
        g = StaticGraph.StaticGraph(self.filename)
        g.set_timeframe(a, b)

        # initialize vertex set and edgelist
        g.vertexList = self.vertexList
        for vertex in self.vertexList.values():
            g.edgeList[vertex] = []

        # iterate over all vertices and check whether an edge for vertex pair (u,v) exists for the entire interval [a,b]
        for vertex in self.vertexList.values():
            tmp = {}
            if a in self.edgeList[vertex].keys():
                for neighbor in self.edgeList[vertex][a]:
                    tmp[neighbor] = 1
                curr_time = a + 1
                while curr_time <= b:
                    try:
                        for neighbor in self.edgeList[vertex][curr_time]:
                            if neighbor in tmp.keys():
                                tmp[neighbor] += 1
                        curr_time += 1
                    except KeyError:
                        curr_time += 1
                        continue
                for neighbor in tmp:
                    if tmp[neighbor] == lifetime:
                        g.edgeList[vertex].append(neighbor)
                        g.numEdges += 1

        # update variables and return graph
        g.numEdges = g.numEdges/2
        g.name_space = self.name_space
        return g

    # returns the strict positive neighborhood of vertex (all vertices with id higher than the vertex [pivot] itself)
    def strict_positive_neighborhood(self, vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]

        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id > vertex.id:
                tmp[neighbor.id] = neighbor
        return tmp

    # returns the inclusive positive neighborhood of vertex (including the vertex)
    def inclusive_positive_neighborhood(self, vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]
        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id > vertex.id:
                tmp[neighbor.id] = neighbor
        tmp[vertex.id] = vertex
        return tmp

    # returns the strict negative neighborhood of vertex (all vertices with id lower than the vertex [pivot] itself)
    def strict_negative_neighborhood(self,vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]
        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id < vertex.id:
                tmp[neighbor.id] = neighbor
        return tmp

    # returns the inclusive negative neighborhood of vertex (including the vertex)
    def inclusive_negative_neighborhood(self, vertex):
        # type: (Vertex) -> collections.OrderedDict[int, Vertex]
        tmp = collections.OrderedDict()
        for neighbor in self.getNeighbors(vertex):
            if neighbor.id < vertex.id:
                tmp[neighbor.id] = neighbor
        tmp[vertex.id] = vertex
        return tmp

    # create a variation of this graph for a given delta (see method on creating delta graph on paper)
    def create_delta_graph(self, delta):

        # it is the same graph for delta = 0
        if delta == 0:
            return self

        # crop delta to reasonable interval
        if delta > self.t_max:
            delta = self.t_max
        if delta < 0:
            delta = 0

        # create graph
        g = TempGraph(self.filename + str(delta))
        g.delta = delta
        num_edges = 0
        g.vertexList = self.vertexList
        for vertex in self.vertexList.values():
            g.edgeList[vertex] = collections.OrderedDict()

        # iterate over timestamps and create edges for all timestamps according to delta
        for timestamp in range(self.t_min, self.t_max + 1 - delta):
            for vertex in self.vertexList.values():
                for t in range(0, delta+1):
                    if timestamp + t in self.edgeList[vertex]:
                        for neighbor in self.edgeList[vertex][timestamp + t]:
                            if timestamp not in g.edgeList[vertex]:
                                g.edgeList[vertex][timestamp] = []
                            if timestamp not in g.edgeList[neighbor]:
                                g.edgeList[neighbor][timestamp] = []
                            if neighbor not in g.edgeList[vertex][timestamp]:
                                g.edgeList[vertex][timestamp].append(neighbor)
                                g.edgeList[neighbor][timestamp].append(vertex)
                                num_edges += 1

        # set variables
        g.t_max = self.t_max - delta
        g.t_min = self.t_min
        g.set_timespace(self.time_space)
        g.set_namespace(self.name_space)
        g.numVertices = self.numVertices
        g.numEdges = num_edges

        return g


    def __iter__(self):
        return iter(self.vertexList.values())

    def set_timespace(self, timespace):
        self.time_space = timespace

    def set_namespace(self, namespace):
        self.name_space = namespace

    # Prints for graph data:

    def print_values(self):
        print("n: " + str(self.numVertices))
        print("m: " + str(self.numEdges))
        print("t_min: " + str(self.t_min))
        print("t_max: " + str(self.t_max))


    def print_graph(self):
        print("TempGraph (" + self.filename + "):")
        print("Vertices:")
        print(self.printVertices())
        print("Edges:")
        print(self.printEdges())
        print("t_min: " + str(self.t_min))
        print("t_max: " + str(self.t_max))
        print("Namespace: ")
        print(self.name_space)
        print("Timespace:")
        print(self.time_space)












