from StaticGraph import *
from TempGraph import *
from isolated_subsets import *
from Vertex import Vertex
import collections
import numpy as np
from utils import *
import math
from server_tests import *

# Enumeration Stage consists of finding isolated cliques via a vertex cover subroutine
# Then for each of we compute all isolated subsets depending on the subset variation
def enumeration_stage(temp_graph, static_graph, vertex, clique, c):
    # type: (TempGraph, StaticGraph, Vertex, collections.OrderedDict[int, Vertex], float) -> list[collections.OrderedDict[int, Vertex]]

    # compute a few needed parameters
    degree = len(static_graph.edgeList[vertex])
    d = degree + 1 - len(clique)
    c_prim = math.ceil(c - d) - 1

    # Create complement graph for vertex cover subroutine
    complement_graph = static_graph.complement_graph(clique, vertex.name)
    complement_graph.sort_by_degree()

    # minimal vertex cover of size c_prim from complement graph G[C]
    all_covers = vertex_cover(complement_graph, c_prim)

    # compute the list of all candidate cliques for isolated cliques
    iso_cliques = []  # type: list[collections.OrderedDict[int, Vertex]]
    for cover in all_covers:
        if temp_graph.del_small_clique >= (len(clique) - len(cover)):
            continue
        tmp = clique.copy()
        for v in cover:
            tmp.pop(v.name)
        iso_cliques.append(tmp.copy())

    # run the isolated subset method to extract all isolated cliques from iso_cliques depending on the used subset method
    result = []  # type: list[collections.OrderedDict[int, Vertex]]
    for current_clique in iso_cliques:
        temp_graph.num_candidate_sets += 1
        remaining_cliques = isolated_subset(current_clique, temp_graph, static_graph, c, degree)
        for remaining_clique in remaining_cliques:
            temp_graph.num_clique_before_time_max += 1
            if temp_graph.del_small_clique >= len(remaining_clique):
                continue
            if len(remaining_clique) != 0:
                result.append(remaining_clique)

    return result


# Vertex Cover subroutine
# We compute the vertex cover on the complement graph to find potential isolated cliques
def vertex_cover(g, k):
    # type: (StaticGraph, float) -> list[set]

    if k < 0:
        return []

    # Here we create the adjacency matrix and list with degrees
    edge_matrix = np.zeros([g.numVertices, g.numVertices])
    degree_list = [0] * g.numVertices

    for vertex in g.vertexList.values():
        degree_list[vertex.id] = g.getDegree(vertex)
        for neighbor in g.edgeList[vertex]:
            edge_matrix[vertex.id][neighbor.id] = 1
            edge_matrix[neighbor.id][vertex.id] = 1

    cover = set()
    all_cliques = []  # type: list[set]
    vertex_branch(g, k, cover, edge_matrix, degree_list, all_cliques)
    return all_cliques

# When there are no edges uncovered, we return true
def is_cover(degree_list):
    # type: (list[int]) -> bool
    for degree in degree_list:
        if degree != 0:
            return False
    return True


def add_to_set(g, cover, edge_matrix, degree_list, vertex, backtracking):
    # type: (StaticGraph, set, np.ndarray, list[int], Vertex, set) -> None

    cover.add(vertex)
    # Vertex has no neighbors anymore
    degree_list[vertex.id] = 0
    # For every previous neighbor
    for neighbor in g.edgeList[vertex]:
        # if they share an edge currently [could be removed in previous steps]
        if edge_matrix[vertex.id][neighbor.id]:
            # remove the edge and reduce the degree of neighbor by 1
            edge_matrix[vertex.id][neighbor.id] = 0
            edge_matrix[neighbor.id][vertex.id] = 0
            degree_list[neighbor.id] -= 1
            backtracking.add((vertex.id, neighbor.id))


def check_minimal(g, cover):

    for vertex in cover:
        necessary = False
        for neighbor in g.edgeList[vertex]:
            if not cover.__contains__(neighbor):
                necessary = True
        if not necessary:
            return False
    return True


def vertex_branch(g, k, cover, edge_matrix, degree_list, all_cliques):
    # type: (StaticGraph, float, set, np.ndarray, list, list[set]) -> bool

    if k < 0:
        return False
    if is_cover(degree_list):
        if check_minimal(g, cover):
            all_cliques.append(cover.copy())
            return True
        return False

    # Take vertex with highest degree left
    max_deg = g.numVertices - 1
    for curr in range(g.numVertices - 1, 0, -1):
        if degree_list[curr] > degree_list[max_deg]:
            max_deg = curr

    vertex = g.vertexList[max_deg]
    backtracking = set()

    # Case1
    add_to_set(g, cover, edge_matrix, degree_list, vertex, backtracking)
    vertex_branch(g, k - 1, cover, edge_matrix, degree_list, all_cliques)

    cover.remove(vertex)
    for edge in backtracking:
        edge_matrix[edge[0]][edge[1]] = 1
        edge_matrix[edge[1]][edge[0]] = 1
        degree_list[edge[1]] += 1
        degree_list[edge[0]] += 1
    backtracking.clear()

    # Case2
    count = 0
    tmp = set()
    for neighbor in g.edgeList[vertex]:
        if not cover.__contains__(neighbor):
            count += 1
            add_to_set(g, cover, edge_matrix, degree_list, neighbor, backtracking)
            tmp.add(neighbor)
    vertex_branch(g, k - count, cover, edge_matrix, degree_list, all_cliques)
    for neighbor in tmp:
        cover.remove(neighbor)
    for edge in backtracking:
        edge_matrix[edge[0]][edge[1]] = 1
        edge_matrix[edge[1]][edge[0]] = 1
        degree_list[edge[1]] += 1
        degree_list[edge[0]] += 1
    backtracking.clear()


def print_vertex_cover(cover):
    print("{", end='')
    i = 0
    for vertex in cover:
        if i == 0:
            print(str(vertex.name), end='')
        else:
            print(", " + str(vertex.name), end='')
        i += 1
    print("}", end='\n')


def print_vertex_covers(cover):
    # type: (list[set]) -> None
    for vertex_set in cover:
        print_vertex_cover(vertex_set)


def max_id(current_clique):
    highest_id = 0
    if len(current_clique) == 0:
        return None
    for vertex in current_clique.values():
        highest_id = max(highest_id, vertex.id)
    return highest_id




