import collections
from StaticGraph import *
from TempGraph import *
import time
from server_tests import *


# Compute all isolated subsets from the current clique
# Given:
# clique
# temporal graph
# static graph (intersection graph)
# isolation parameter c
# degree of pivot element
def isolated_subset(current_clique, temp_graph, static_graph, c, degree):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, float, int, int) -> list[collections.OrderedDict[int, Vertex]]

    # We compute the set of all isolated cliques extracted from the currently computed clique
    # This clique may be c-isolated according to a given subset variation.
    # Some subset variations can extract multiple cliques out of the initial clique (if not c-isolated at first)
    if temp_graph.mode == 0:
        return isolated_subset_alltime_max(current_clique, temp_graph, static_graph, c, degree)
    elif temp_graph.mode == 1:
        return isolated_subset_max_usually(current_clique, temp_graph, static_graph, c, degree)
    elif temp_graph.mode == 2:
        return isolated_subset_usally_average(current_clique, temp_graph, static_graph, c, degree)
    elif temp_graph.mode == 3:
        return isolated_subset_alltime_average(current_clique, temp_graph, static_graph, c, degree)
    elif temp_graph.mode == 4:
        return isolated_subset_average_alltime(current_clique, temp_graph, static_graph, c, degree)
    elif temp_graph.mode == 5:
        return isolated_subset_usally_max(current_clique, temp_graph, static_graph, c, degree)


def isolated_subset_alltime_max(clique, temp_graph, static_graph, c, degree):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, float, int) -> list[collections.OrderedDict[int, Vertex]]

    # compute s_v (Line 1-2 in alltime-max-isolatedSubsets)
    length = len(clique)
    k = min(int(degree - c + 2), length)
    s = compute_alltime_sv(clique, temp_graph, static_graph.start_time, static_graph.end_time, length)

    # Remove offending vertices until we either success or fail
    while length >= k:
        change = False
        for vertex in s.keys():
            if s[vertex] >= c:
                s.pop(vertex)
                clique.pop(vertex.id)
                length -= 1
                for others in s.keys():
                    s[others] += 1
                change = True
                break
        if not change:
            return [clique]
    return []




def isolated_subset_max_usually(clique, temp_graph, static_graph, c, degree):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, float, int) -> list[collections.OrderedDict[int, Vertex]]

    # Compute s_v
    lifetime = static_graph.end_time - static_graph.start_time + 1
    length = len(clique)
    k = min(int(degree - c + 2), length)
    s = compute_usally_sv(clique, temp_graph, static_graph, length)

    # Remove offending vertices until we either success or fail
    while length >= k:
        change = False
        for vertex in s.keys():
            if s[vertex] >= c * lifetime:
                s.pop(vertex)
                clique.pop(vertex.id)
                length -= 1
                for others in s.keys():
                    s[others] += lifetime
                change = True
                break
        if not change:
            return [clique]
    return []


def isolated_subset_usally_average(clique, temp_graph, static_graph, c, degree):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, float, int) -> list[collections.OrderedDict[int, Vertex]]

    # Compute s_v
    clique_sets = []  # type: list[collections.OrderedDict[int, Vertex]]
    length = len(clique)
    s = compute_usally_sv(clique, temp_graph, static_graph, length)
    lifetime = static_graph.end_time - static_graph.start_time + 1

    # Map vertex id's to 1,...,k for later procedure sorted by s_v value
    new_ranking, reverse_mapping = map_clique_to_ranking(clique, s)

    # Slightly different approach from paper. We compute the score in such a way, that the condition is fullfilled
    # if e_c < 0
    e_c = 0
    for vertex in clique.values():
        e_c += s[vertex]
    e_c -= c*lifetime*len(clique)
    if e_c < 0:     # NEEDS to be smaller than 0, because sum s_v - c(b+1-a) < 0, then usally_avg condition fullfilled
        clique_sets.append(clique.copy())
        return clique_sets

    d = min(int(length - degree + c - 2), length)

    # Recursivally call the usually_average method since we need to iterate over a lot of subsets of the current
    # (non-c-isolated) clique and check for these subsets wheter they are c-isoalted
    recursion_usally_average(clique.copy(), -1, clique_sets, new_ranking, reverse_mapping, s, e_c, length, c,
                             lifetime, d)

    return clique_sets


def recursion_usally_average(current_clique, removed_vertex, clique_sets, new_ranking, reverse_mapping, s, e_c, k, c, lifetime, ttl):
    # type: (collections.OrderedDict[int, Vertex], int, list[collections.OrderedDict[int, Vertex]], collections.OrderedDict[Vertex, int], collections.OrderedDict[int, Vertex], collections.OrderedDict[Vertex, int], float, int, float, int, int) -> None

    # We only allow a specific recursion depth. If time-to-live exhausted, just break
    if ttl < 0:
        return

    # If removed-vertex is -1, this is our first iteration. Just check all variations of removing a single vertex
    # Note, we remove vertices in such a way, that once we remove a vertex v with a specific id, we only remove vertices
    # with id smaller then v afterwards to remove doubles
    if removed_vertex == -1:
        for iteration in range(1, min(len(current_clique), int(c))):
            tmp = current_clique.copy()
            tmp.pop(reverse_mapping[iteration].id)

            # substract s_v of removed vertex and add lifetime times num remaining vertices to e_c
            tmp_e_c = e_c - s[reverse_mapping[iteration]] + len(tmp)*lifetime
            recursion_usally_average(tmp, iteration, clique_sets, new_ranking, reverse_mapping, s,
                                     tmp_e_c, k, c, lifetime, ttl - 1)
        return
    if e_c < 0:
        clique_sets.append(current_clique.copy())
        return
    # Remove a smaller vertex and update the score e_c
    for iteration in range(removed_vertex + 1, min(k, int(c))):
        tmp = current_clique.copy()
        tmp.pop(reverse_mapping[iteration].id)
        tmp_e_c = e_c - s[reverse_mapping[iteration]] + len(tmp) * lifetime
        recursion_usally_average(tmp, iteration, clique_sets, new_ranking, reverse_mapping, s, tmp_e_c, k, c,
                                 lifetime, ttl - 1)
    return


def isolated_subset_alltime_average(clique, temp_graph, static_graph, c, degree):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, float, int) -> list[collections.OrderedDict[int, Vertex]]

    # Compute v_i_k storing the vertices for all timestamps in order of degree
    length = len(clique)
    d = []  # type: list[int]
    ttl = min(int(length - degree + c - 2),length)
    clique_sets = []  # type: list[collections.OrderedDict[int, Vertex]]
    v_i_k = compute_alltime_average_sv(clique, temp_graph, static_graph, length)   # type: list[list[Vertex]]

    # Recursivally call the alltime_average method since we need to iterate over a lot of subsets of the current
    # (non-c-isolated) clique and check for these subsets wheter they are c-isoalted
    recursion_alltime_average(clique.copy(), d, clique_sets, v_i_k, temp_graph, c, ttl, static_graph.start_time,
                              static_graph.end_time, ttl + 1)
    return clique_sets



def smallest_not_alltime_average_isolated(current_clique, temp_graph, c, a, b):
    # return the smallest value i s.t. current_clique not average isolated in static graph h_i for i in start/end time
    # avg-isolated: sum outdegree < c*|C|
    threshold = c * len(current_clique)
    length = len(current_clique)

    i = a
    while i <= b:
        outdegree = 0
        for vertex in current_clique.values():
            outdegree += len(temp_graph.edgeList[vertex][i]) + 1 - length
            if outdegree >= threshold:
                return b - i
        i += 1
    return -1



def recursion_alltime_average(clique, d, clique_sets, v_i_k, temp_graph, c, k, a, b, ttl):
    # type: (collections.OrderedDict[int, Vertex], list[int], list[collections.OrderedDict[int, Vertex]], list[list[Vertex]], TempGraph, float, int, int, int, int) -> None

    # We only allow a specific recursion depth. If time-to-live exhausted, just break
    if ttl < 0:
        return

    # find smallest i such that clique not_avg_isolated in temp_graph [if exists]
    i = smallest_not_alltime_average_isolated(clique, temp_graph, c, a, b)
    if i == -1:
        clique_sets.append(clique.copy())
        return
    j = 0
    for position in range(0, len(v_i_k[i])):
        if d.__contains__(v_i_k[i][position]):
            j = max(j, position)

    # Iterate over possibilies of vertices to remove
    for iteration in range(j, k):
        tmp = clique.copy()
        if tmp.__contains__(v_i_k[i][iteration].id):
            tmp.pop(v_i_k[i][iteration].id)
            d_copy = d.copy()
            d_copy.append(v_i_k[i][iteration].id)
            recursion_alltime_average(tmp, d_copy, clique_sets, v_i_k, temp_graph, c, k, a, b, ttl - 1)
    return

# returns a list of the vertices of clique sorted by outdegree for current timestamp in descending order
def compute_alltime_average_sv(current_clique, temp_graph, static_graph, length):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, int) -> list[list[Vertex]]

    s = []              # type list[list[Vertex]]
    curr_timestamp = static_graph.start_time
    while curr_timestamp <= static_graph.end_time:
        tmp = sorted(current_clique.values(),
                     key=lambda vertex: len(temp_graph.edgeList[vertex][curr_timestamp]) - length + 1,
                     reverse=True)  # type: list[Vertex]
        s.append(tmp.copy())
        curr_timestamp += 1
    return s


def isolated_subset_average_alltime(clique, temp_graph, static_graph, c, degree):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, float, int) -> list[collections.OrderedDict[int, Vertex]]

    clique_sets = []         # type: list[collections.OrderedDict[int, Vertex]]
    length = len(clique)
    ttl = min(int(length - degree + c - 2), length)

    # Compute s_v
    s = compute_average_alltime_sv(clique, temp_graph, static_graph, length)

    # Map vertex id's to 1,...,k for later procedure sorted by s_v value
    new_ranking, reverse_mapping = map_clique_to_ranking(clique, s)

    e_c = 0
    for vertex in clique.values():
        e_c += s[vertex]
    e_c -= c * length

    if e_c < 0:
        clique_sets.append(clique.copy())
        return clique_sets

    recursion_average_alltime(clique, -1, clique_sets, new_ranking, reverse_mapping, s, e_c, ttl, c, ttl + 1)

    return clique_sets

def recursion_average_alltime(clique, removed_vertex, clique_sets, new_ranking, reverse_mapping, s, e_c, k, c, ttl):
    # type: (collections.OrderedDict[int, Vertex], int, list[collections.OrderedDict[int, Vertex]], collections.OrderedDict[Vertex, int], collections.OrderedDict[int, Vertex], collections.OrderedDict[Vertex, int], float, int, float, int) -> None

    # We only allow a specific recursion depth. If time-to-live exhausted, just break
    if ttl < 0:
        return

    # If removed-vertex is -1, this is our first iteration. Just check all variations of removing a single vertex
    # Note, we remove vertices in such a way, that once we remove a vertex v with a specific id, we only remove vertices
    # with id smaller then v afterwards to remove doubles
    if removed_vertex == -1:
        for iteration in range(1, min(len(clique), int(c))):
            tmp = clique.copy()
            tmp.pop(reverse_mapping[iteration].id)

            # substract s_v of removed vertex and add 1 for each remaining vertex, 'cuz s_v of all other vertices
            # increases by 1.
            tmp_e_c = e_c - s[reverse_mapping[iteration]] + len(tmp)
            recursion_average_alltime(tmp, iteration, clique_sets, new_ranking, reverse_mapping, s,
                                     tmp_e_c, k, c,  ttl - 1)
        return

    if e_c < 0:
        clique_sets.append(clique.copy())
        return
    # Iterate over possibilies of vertices to remove
    for iteration in range(max(1, k - int(c)), removed_vertex - 1):
        tmp = clique.copy()
        tmp.pop(reverse_mapping[iteration].id)
        tmp_e_c = e_c - s[reverse_mapping[iteration]] + len(tmp)
        recursion_average_alltime(tmp, iteration, clique_sets, new_ranking, reverse_mapping, s, tmp_e_c, k, c, ttl - 1)
    return



def isolated_subset_usally_max(current_clique, temp_graph, static_graph, c, degree):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, float, int) -> list[collections.OrderedDict[int, Vertex]]
    # TODO: Not yet part of the paper, thus no code available yet
    return []


def compute_average_alltime_sv(clique, temp_graph, static_graph, length):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, int) -> collections.OrderedDict[Vertex, int]
    s = collections.OrderedDict()  # type: collections.OrderedDict[Vertex, int]
    for vertex in clique.values():
        max_outgoing = 0
        curr_timestamp = static_graph.start_time
        while curr_timestamp <= static_graph.end_time:
            max_outgoing = max(max_outgoing, len(temp_graph.edgeList[vertex][curr_timestamp]))
            curr_timestamp += 1
        s[vertex] = max_outgoing - length + 1       # add - length + 1 add the end instead of every time
    return s

def compute_usally_sv(clique, temp_graph, static_graph, length):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, StaticGraph, int) -> collections.OrderedDict[Vertex, int]
    s = collections.OrderedDict()  # type: collections.OrderedDict[Vertex, int]
    for vertex in clique.values():
        most_neighbors_for_timestamp = 0
        curr_timestamp = static_graph.start_time
        while curr_timestamp <= static_graph.end_time:
            most_neighbors_for_timestamp += len(temp_graph.edgeList[vertex][curr_timestamp]) - length + 1
            curr_timestamp += 1
        s[vertex] = most_neighbors_for_timestamp
    return s


def compute_alltime_sv(clique, temp_graph, a, b, length):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, int, int, int) -> collections.OrderedDict[Vertex, int]
    # Compute highest number of outgoing edges for each vertex for one timestamp
    s = collections.OrderedDict()  # type: collections.OrderedDict[Vertex, int]
    for vertex in clique.values():
        most_neighbors_for_timestamp = 0
        curr_timestamp = a
        while curr_timestamp <= b:
            cur = len(temp_graph.edgeList[vertex][curr_timestamp])
            most_neighbors_for_timestamp = max(cur, most_neighbors_for_timestamp)
            curr_timestamp += 1
        s[vertex] = most_neighbors_for_timestamp + 1 - length   # add this once in the end instead of every time
    return s


def map_clique_to_ranking(current_clique, s):
    # type: (collections.OrderedDict[int, Vertex], collections.OrderedDict[Vertex, int]) -> (collections.OrderedDict[Vertex, int], collections.OrderedDict[int, Vertex])
    tmp = sorted(current_clique.values(), key=lambda curr_vertex: s[curr_vertex])  # type: list[Vertex]
    new_ranking = collections.OrderedDict()  # type: collections.OrderedDict[Vertex, int]
    reverse_mapping = collections.OrderedDict()  # type: collections.OrderedDict[int, Vertex]
    increment = 1
    for vertex in tmp:
        new_ranking[vertex] = increment
        reverse_mapping[increment] = vertex
        increment += 1
    return new_ranking, reverse_mapping































