import argparse
import sys
import time
import trimming
import enumeration
from StaticGraph import *
from TempGraph import TempGraph
from utils import *
import collections
from isolated_subsets import *
import datetime
from server_tests import *



def main():
    parser = argparse.ArgumentParser(description="Enumerates temporal isolated cliques in a given temporal graph.",add_help=False)
    parser.add_argument("-d", "--delta", type=int, default=1, help="Delta, default is 1.")
    parser.add_argument("-c", "--c",  default=5, help="Input Parameter c, default is 5.")
    parser.add_argument("-n", "--filename", default="", help="file name of data.")
    parser.add_argument("-m", "--mode", type=str, default="0", help="index of subset version to be used")
    parser.add_argument("-o", "--output", action="store_true", help="Enable additional output")
    parser.add_argument("-t", "--timestamp", type=int, default=-1, help="number of timestamps to run algorithm on. -1 for all")
    parser.add_argument("-k", "--del_small_clique", type=int, metavar="X", default=1, help="Ignore all cliques with at most X vertices (default: 1)")
    parser.add_argument("-v", "--version", type=int, default=1, help="Set to 1 for new vertex maximality check (default). Set to 0 for old version.")
    parser.add_argument("-p", "--period", type=int, default=20,
                        help="time between two timestamps in seconds. Used to identify non-consecutive timestamps")
    parser.add_argument("-h", "--help", action="store_true", help="Show helpful commands")

    # Preperate parsing:
    args = parser.parse_args()
    if args.help:
        print_help_message()
        return


    output = False
    if args.output:
        output = True

    if not len(args.filename):
            print("We need an input file. Use --filename \"[path_to_file]\".")
            print("Use --help for assistance.")
            return

    now = datetime.datetime.now()
    outputname = "comparison_" + str(now.strftime("%Y-%m-%d")) + ".csv"
    mode = args.mode
    if not str.isdigit(mode):  # In this case we have the string version -> translate it to integer
        mode = reverse_map_mode(mode)
    else:
        mode = int(args.mode)

    # Call program
    algorithm_clipped(args.filename, args.timestamp, float(args.c), mode, int(args.delta), outputname, output, int(args.period), int(args.version))


def print_help_message():

    print("Example on how to use the program:")
    print("python3 temp_iso_cliques.py --filename [file] -m [0-4]")
    print("------------------------------------------------")
    print("Possible parameters:")
    print("        -d,--delta           :     Increase for denser graphs (layering of time edges)")
    print("        -c,--c               :     Increase c to allow less c-isolated cliques")
    print("[N]     -n,--filename        :     Input Filename after.")
    print("        -m,--mode            :     Indicates the isolation type to use. Default is 0")
    print("            Possible isolation types (Both int and string variant work as input):")
    print("                0, alltime_max")
    print("                1, max_usually")
    print("                2, usally_average")
    print("                3, alltime_average")
    print("                4, average_alltime")
    print("                5, usally_max     (not implemented yet)")
    print("        -o,--output          :     Enables additional output while computation to check progress.")
    print("        -t,--timestamp       :     Default at -1. Use positiv integer to restrict computation to first [x] time layers.")
    print("        -k,--del_small_clique:     Ignore all cliques with at most [x] many vertices. Default is at 1 to ignore single vertices as cliques.")
    print("        -v,--version         :     Set to 1 for new vertex maximality check. Use 0 for old version (brute force). Default at 1.")
    print("        -p,--period          :     Set the interval (period) between timestamps to accurately compute cliques over intervals. Default at 20.")
    print("        -h,--help            :     Print this text. Don't use this parameter when computing cliques since it stops the program.")
    print("----------------------------------")
    print("Parameters marked with [N] are necessarry input.")



# Returns true if clique exists for timestamp on temporal graph (not necessarily isolated)
def clique_exists_for_timestamp(clique, tempgraph, a_prim):
    # type: (collections.OrderedDict[int, Vertex], TempGraph, int) -> bool

    for u in clique.values():
        for v in clique.values():
            if u.name != v.name:
                if a_prim not in tempgraph.edgeList[u].keys():
                    return False
                if not tempgraph.edgeList[u][a_prim].__contains__(v):
                    return False
    return True


# Compare the already computed list for a timestamp with the given clique and check if it
# is maximal (no clique in time_max_cliques is better / at least as good)
def is_max_for_timestamp(time_max_cliques, clique):
    # type: (list[collections.OrderedDict[int, Vertex]], collections.OrderedDict[int, Vertex]) -> bool

    keys = set([vertex.name for vertex in clique.values()])
    for max_clique in time_max_cliques:
        if set([vertex.name for vertex in max_clique.values()]) >= keys:
            return False

    return True


# Returns true if clique is vertex maximal for intervall [a,b], c-isolated and time maximal for the usually variant
def is_maximal_usually(tempgraph, static_graph, clique, a, b, c):

    if a == tempgraph.t_min and b == tempgraph.t_max:
        return is_vertex_maximal(tempgraph, static_graph, clique, a, b, c)
    max_b = tempgraph.t_max + 1
    a_prim = a
    if b == tempgraph.t_max:
        a_prim -= 1
    while a_prim > -1:
        if not clique_exists_for_timestamp(clique, tempgraph, a_prim):
            break
        b_prim = b
        if a_prim == a and b != tempgraph.t_max:
            b_prim += 1
        while b_prim < max_b:
            if max_b == tempgraph.t_max + 1:
                if not clique_exists_for_timestamp(clique, tempgraph, b_prim):
                    max_b = b_prim
                    break
            if is_c_isolated_usually(tempgraph, clique, a_prim, b_prim, c) or \
                not is_vertex_maximal(tempgraph, static_graph, clique, a_prim, b_prim, c):
                return False
            b_prim += 1
        a_prim -= 1
    return True


# Note: This is the old version.
# Checks for all cliques found for intervall [a,b] whether they are vertex maximal and time maximal.
def time_max_cliques_usually_old_version(tempgraph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c):
    # type: (TempGraph, StaticGraph, collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]], list[collections.OrderedDict[int, Vertex]], int, int, float) -> None

    # Use a brute-force check between all computed cliques to check if one is a subset of another
    vertex_max_cliques = vertex_maximal_cliques(vertex_max_cliques)
    # If first time interval, we just add them
    if a == 0 and b == 0:
        # first ever timestamps
        for clique in vertex_max_cliques:
            time_max_cliques[a][b].append(clique)
        return

    # Check all smaller time intervalls and remove cliques that are subsets of our newly found cliques
    length = len(vertex_max_cliques)
    i = 0
    while i < length:
        clique = vertex_max_cliques[i]
        better_then_old_clique = False
        for b_prim in range(a, b):
            j = 0
            length2 = len(time_max_cliques[a][b_prim])
            while j < length2:
                old_clique = time_max_cliques[a][b_prim][j]
                keys_new = set([x.name for x in clique.values()])
                keys_old = set([x.name for x in old_clique.values()])
                if keys_new >= keys_old:
                    better_then_old_clique = True
                    time_max_cliques[a][b_prim].remove(old_clique)
                    length2 -= 1
                j += 1
        if better_then_old_clique:
            time_max_cliques[a][b].append(clique)
            vertex_max_cliques.remove(clique)
            length -= 1
            continue
        i += 1

    # Remove all cliques from being potentially put to time_max_cliques if they are not time maximal
    length = len(vertex_max_cliques)
    i = 0
    while i < length:
        clique = vertex_max_cliques[i]
        max_b = tempgraph.t_max + 1
        a_prim = a - 1
        while a_prim > -1:
            if not clique_exists_for_timestamp(clique, tempgraph, a_prim):
                break
            b_prim = b
            while b_prim < max_b:
                if max_b == tempgraph.t_max + 1:
                    if not clique_exists_for_timestamp(clique, tempgraph, b_prim):
                        max_b = b_prim
                        break
                if not is_max_for_timestamp(time_max_cliques[a_prim][b_prim], clique):
                    a_prim = -1
                    b_prim = max_b
                    vertex_max_cliques.remove(clique)
                    length -= 1
                    i -= 1
                b_prim += 1
            a_prim -= 1

        i += 1

# Add the cliques found for intervall [a,b] to the set of time maximal cliques for the usually variations
def add_to_time_max_cliques_usally(tempgraph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c):
    # type: (TempGraph, StaticGraph, collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]], list[collections.OrderedDict[int, Vertex]], int, int) -> None

    # Brute-Force variant (old version)
    if tempgraph.maximality_variant == 0 or tempgraph.mode == 5:
        time_max_cliques_usually_old_version(tempgraph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c)

    # New version (doesn't work for variant 5,usally_max)
    if tempgraph.maximality_variant == 1 and tempgraph.mode != 5:
        length = len(vertex_max_cliques)
        i = 0
        while i < length:
            clique = vertex_max_cliques[i]
            if not is_maximal_usually(tempgraph, static_graph, clique, a, b, c):
                vertex_max_cliques.remove(clique)
                length -= 1
                i -= 1
            i += 1
    # Add the cliques afte all non-maximal ones were removed
    for clique in vertex_max_cliques:
        time_max_cliques[a][b].append(clique)

# Add cliques to set of time maximal cliques (split into different functions for alltime/usually variations
def add_to_time_max_cliques(tempgraph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c):
    if is_alltime(tempgraph.mode):
        add_to_time_max_cliques_alltime(tempgraph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c)
    else:
        add_to_time_max_cliques_usally(tempgraph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c)

# Returns whether clique is c-isolated for intervall [a,b] and variation usually_average
def is_c_isolated_usually_avg(tempgraph,clique, a, b, c):
    # type: (TempGraph, collections.OrderedDict[int, Vertex], int, int, float) -> bool
    len_clique = len(clique)
    outdegree = 0
    for vertex in clique.values():
        curr_timestamp = a
        while curr_timestamp <= b:
            try:
                outdegree += len(tempgraph.edgeList[vertex][curr_timestamp]) + 1 - len_clique
            except KeyError:
                curr_timestamp += 1
                continue
            curr_timestamp += 1
    # return True if usually-average isolated
    return outdegree < c*len_clique*(b + 1 - a)

# Returns whether clique is c-isolated for intervall [a,b] and variation max_usually
def is_c_isolated_max_usually(tempgraph, clique, a, b, c):
    # type: (TempGraph, collections.OrderedDict[int, Vertex], int, int, float) -> bool

    # vertex with highest outdegree:
    len_clique = len(clique)
    max_outdegree = -1

    for vertex in clique.values():
        outdegree = 0
        curr_timestamp = a
        while curr_timestamp <= b:
            try:
                outdegree += len(tempgraph.edgeList[vertex][curr_timestamp]) + 1 - len_clique
            except KeyError:
                curr_timestamp += 1
                continue
            curr_timestamp += 1
        max_outdegree = max(max_outdegree, outdegree)
    # Return True if c-isolated
    return max_outdegree < c*(b + 1 -a)

# Returns whether clique is c-isolated for intervall [a,b] and variations alltime_max
def is_c_isolated_alltime_max(temp_graph, clique, t, c):
    # Compute whether it is alltime max isolated for timestamp t and isolation parameter c
    len_clique = len(clique)
    most_neighbors_for_timestamp = 0
    for vertex in clique.values():
        cur = len(temp_graph.edgeList[vertex][t]) + 1 - len_clique
        most_neighbors_for_timestamp = max(cur, most_neighbors_for_timestamp)

    if most_neighbors_for_timestamp >= c:
        return False
    return True

# Returns whether clique is c-isolated for intervall [a,b] and variations alltime_average
def is_c_isolated_alltime_average(temp_graph, clique, t, c):
    # Compute whether it is alltime average isolated for timestamp t and isolator c
    len_clique = len(clique)
    sum_over_outdegree = 0
    for vertex in clique.values():
        sum_over_outdegree += len(temp_graph.edgeList[vertex][t]) + 1 - len_clique
    if sum_over_outdegree >= c * len_clique:
        return False
    return True

# Returns whether clique is c-isolated for intervall [a,b] and variations average_alltime
def is_c_isolated_average_alltime(temp_graph, clique, a, b, c):
    len_clique = len(clique)
    overall_outdegree = 0
    for vertex in clique.values():
        vertex_outdegree = 0
        cur_time = a
        while cur_time <= b:
            try:
                vertex_outdegree = max(vertex_outdegree, len(temp_graph.edgeList[vertex][cur_time]) + 1 - len_clique)
            except KeyError:
                continue
            cur_time += 1
        overall_outdegree += vertex_outdegree
    if overall_outdegree >= c * len_clique:
        return False
    return True


# Returns whether clique is c-isolated for intervall [a,b] (Split between usually variations)
# Note: usally_max not added because it is not implemented yet.
def is_c_isolated_usually(tempgraph, clique, a, b, c):

    if tempgraph.mode == 1:
        return is_c_isolated_max_usually(tempgraph, clique, a, b, c)
    elif tempgraph.mode == 2:
        return is_c_isolated_usually_avg(tempgraph, clique, a, b, c)
    return False

# Returns whether clique is c-isolated for intervall [a,b] (Split between alltime variations)
def is_c_isolated_alltime(temp_graph, clique, timestamp, a, b, c):
    # type: (TempGraph, collections.OrderedDict[int, Vertex], int, int, int, float) -> bool

    # checks for given alltime variation if the clique is c-isoalted for the new timestamp "timestamp"
    # when we observe the intervall [a,b] first. Timestamp is usually a-1 or b+1.
    if temp_graph.mode == 0:
        # Compute highest number of outgoing edges for each vertex for one timestamp
        return is_c_isolated_alltime_max(temp_graph, clique, timestamp, c)
    elif temp_graph.mode == 3:
        # Compute whether sum over outdegree is bigger than c*|C|
        return is_c_isolated_alltime_average(temp_graph, clique, timestamp, c)
    elif temp_graph.mode == 4:
        # sum over max outdegree for each vertex in clique. Shall be smaller than c*|C|
        return is_c_isolated_average_alltime(temp_graph, clique, a, b, c)
    return False

# Returns whether clique found for interval [a,b] exists and is isolated for intervall [a-1,b] and [a,b+1]
def is_isolated_alltime(tempgraph, clique, a, b, c):
    # type: (TempGraph, collections.OrderedDict[int, Vertex], int, int, float) -> bool
    if a != tempgraph.t_min and \
            clique_exists_for_timestamp(clique, tempgraph, a - 1) and \
            is_c_isolated_alltime(tempgraph, clique, a - 1, a - 1, b, c):
        return False
    if b != tempgraph.t_max and \
            clique_exists_for_timestamp(clique, tempgraph, b + 1) and \
            is_c_isolated_alltime(tempgraph, clique, b + 1, a, b + 1, c):
        return False
    return True

# Adds all found cliques for intervall [a,b] to the set of time max cliques if they are time and vertex maximal
def add_to_time_max_cliques_alltime(tempgraph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c):
    # type: (TempGraph, StaticGraph, collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]], list[collections.OrderedDict[int, Vertex]], int, int, float) -> None

    # Brute-Force variant
    if tempgraph.maximality_variant == 0 or tempgraph.mode == 3:
        vertex_max_cliques = vertex_maximal_cliques(vertex_max_cliques)
    # remove clique from new vertex max clique set if we have shown, that its still isolated for a bigger intervall
    length = len(vertex_max_cliques)
    i = 0
    while i < length:
        clique = vertex_max_cliques[i]
        if not is_isolated_alltime(tempgraph, clique, a, b, c):
            vertex_max_cliques.remove(clique)
            length -= 1
            i -= 1
        i += 1

    if tempgraph.maximality_variant == 1 and tempgraph.mode != 3:
        i = 0
        while i < length:
            clique = vertex_max_cliques[i]
            if not is_vertex_maximal(tempgraph, static_graph, clique, a, b, c):
                vertex_max_cliques.remove(clique)
                i -= 1
                length -= 1
            i += 1
    # Add remaining cliques since they are maximal
    for clique in vertex_max_cliques:
        time_max_cliques[a][b].append(clique)


# Bron-Kerbosch recursive part
def static_bronkerbosch_recur(r, p, x, d, static_graph):

    if len(p) == 0 and len(x) == 0:
        clique = collections.OrderedDict()  # type: collections.OrderedDict[int, Vertex]
        for vertex in r:
            clique[vertex.id] = vertex
        if len(clique) > 0:
            d.append(clique)

    i = 0
    length = len(p)
    while i < length:
        vertex = p[i]
        r_prim = r.copy()
        r_prim.append(vertex)
        p_prim = [element for element in p if static_graph.is_neighbor(element, vertex)]
        x_prim = [element for element in x if static_graph.is_neighbor(element, vertex)]

        static_bronkerbosch_recur(r_prim, p_prim, x_prim, d, static_graph)
        p.pop(0)
        length -= 1
        if not x.__contains__(vertex):
            x.append(vertex)

# Bron-Kerbosch subroutine on static graph
def static_bronkerbosch(static_graph, neighborhood):

    # start: R empty, X empty, P vertex set
    d = []  # type: list[collections.OrderedDict[int, Vertex]]
    static_bronkerbosch_recur([], neighborhood, [], d, static_graph)
    return d

# Ner version to check wheter clique found in [a,b] is vertex maximal
# Works for:
#   max-usually(0), usually-avg(1), alltime-max(2) ,avg-alltime(4)
def is_vertex_maximal(tempgraph, static_graph, clique, a, b, c):
    # type: (Tempgraph, StaticGraph, collections.OrderedDict[int, Vertex], int, int, float) -> bool

    # compute pivot element
    for vertex in clique.values():
        pivot = vertex
        break
    min_degree = len(static_graph.getNeighbors(pivot))
    for vertex in clique.values():
        new_deg = static_graph.getDegree(vertex)
        if new_deg < min_degree:
            pivot = vertex
            min_degree = new_deg
    # compute neighborhood of pivot
    neighborhood = static_graph.getNeighbors(pivot)

    # Remove all neighbors of pivot which are not connected to all vertices in the clique or are part of the clique itself
    length = len(neighborhood)
    i = 0
    while i < length:
        vertex = neighborhood[i]
        if vertex in clique.values():
            neighborhood.remove(vertex)
            i -= 1
            length -= 1
        else:
            for element in clique.values():
                if not static_graph.edgeList[vertex].__contains__(element):
                    neighborhood.remove(vertex)
                    i -= 1
                    length -= 1
                    break
        i += 1

    # Compute maximal cliques for vertices in neighborhood
    D = static_bronkerbosch(static_graph, neighborhood)

    for max_clique in D:
        D_prim = max_clique.copy()
        new_clique = clique.copy()
        for vertex in D_prim.values():
            new_clique[vertex.id] = vertex
        while len(D_prim.values()) != 0 and not is_c_isolated(new_clique, a, b, tempgraph, static_graph, c):
            max_outdegree = -1
            d = None
            for vertex in D_prim.values():
                outdegree = 0
                curr_timestamp = a
                while curr_timestamp <= b:
                    try:
                        if tempgraph.mode == 0 or tempgraph.mode == 4:
                            outdegree = max(outdegree, len(tempgraph.edgeList[vertex][curr_timestamp]))
                        if tempgraph.mode == 1 or tempgraph.mode == 2:
                            outdegree += len(tempgraph.edgeList[vertex][curr_timestamp])
                    except KeyError:
                        curr_timestamp += 1
                        continue
                    curr_timestamp += 1
                if max_outdegree < outdegree:
                    max_outdegree = outdegree
                    d = vertex
            if d is not None:
                D_prim.pop(d.id)
                new_clique.pop(d.id)
        if len(D_prim.values()) != 0:
            return False

    return True


def is_c_isolated(clique, a, b, tempgraph, static_graph, c):
    # type: (collections.OrderedDict[int, Vertex], int, int, TempGraph, StaticGraph, float) -> bool

    if tempgraph.mode == 0:
        isolated = True
        curr_timestamp = a
        while curr_timestamp <= b and isolated:
            isolated = is_c_isolated_alltime_max(tempgraph, clique, curr_timestamp, c)
            curr_timestamp += 1
        return isolated

    if tempgraph.mode == 4:
        return is_c_isolated_average_alltime(tempgraph, clique, a, b, c)
    if tempgraph.mode == 1:
        return is_c_isolated_max_usually(tempgraph, clique, a, b, c)
    if tempgraph.mode == 2:
        return is_c_isolated_usually_avg(tempgraph, clique, a, b, c)


# Returns whether the current mode is encoding an alltime variation
def is_alltime(mode):
    # type: (int) -> bool
    if mode == 0 or mode == 3 or mode == 4:
        return True
    return False


# Start the algorithm to compute all c-isolated cliques for the graph
def algorithm(temporal_graph, c, max_time=-1, output=False):
    # type: (TempGraph, float, int, bool) -> collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]]

    # We do this, because for previous iterations on the graph we left with different vertex id's depending on
    # the ordering. We get different results since we assume all vertices to have their name in id at first
    for vertex in temporal_graph.vertexList.values():
        vertex.id = vertex.name
    if max_time == -1:
        max_time = temporal_graph.t_max
    time_max_cliques = collections.OrderedDict()    # type: collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]]
    num_iter = 0

    # Iterate over all possible intervals
    for a in range(temporal_graph.t_min, max_time + 1):
        if output:
            if a % 50 == 0 and a != 0:
                print(str(a) + " ", end='', flush=True)
        static_graph = temporal_graph.getStaticGraph(a, a)
        static_graph.sort_by_degree()
        time_max_cliques[a] = collections.OrderedDict()
        for b in range(a, max_time + 1):
            num_iter += 1
            if a != b:
                # Instead of computing the static graph we use the one from before and just add the new timestamp
                static_graph.add_another_timestamp(temporal_graph, b)
                # From here on we can ignore all further intervals since they contain no edges
                if static_graph.numEdges == 0:
                    break
                static_graph.sort_by_degree()

            temporal_graph.num_timestamps_iter += 1
            time_max_cliques[a][b] = []
            # Compute the candidate set of isolated cliques in trimming stage/enumeration stage
            vertex_max_cliques = run_stages(temporal_graph, static_graph, c)
            # If we find no cliques and use an alltime variant, then there are no cliques for a bigger intervall
            # We cannot guarentee this for the usually variations
            if len(vertex_max_cliques) == 0 and is_alltime(temporal_graph.mode):
                break
            add_to_time_max_cliques(temporal_graph, static_graph, time_max_cliques, vertex_max_cliques, a, b, c)

    # Return the set of all c isolated cliques
    return time_max_cliques


# Method to print the set of cliques found for a given time interval readable (mainly for debugging purpose)
def print_vertex_max_cliques(vertex_max_cliques, a, b):
    # type: (list[collections.OrderedDict[int, Vertex]], int, int) -> None
    print(" ")
    print("[" + str(a) + "," + str(b) + "]")
    for clique in vertex_max_cliques:
        print_clique(clique)
    print(" ")

# Run through trimming and enumeration stage.
# We do this for every vertex as we do not count cliques multiple times by using these vertices as pivot element
def run_stages(temporal_graph, static_graph, c):
    # type: (TempGraph, StaticGraph, float) -> list[collections.OrderedDict[int, Vertex]]
    cliques = []  # type: list[collections.OrderedDict[int, Vertex]]
    for vertex in static_graph.vertexList.values():
        clique = trimming.trimming_stage(static_graph, vertex, c, temporal_graph.del_small_clique)
        if len(clique) == 0 or (len(clique) <= temporal_graph.del_small_clique):
            continue
        tmp = enumeration.enumeration_stage(temporal_graph, static_graph, vertex, clique, c)
        for subset in tmp:
            cliques.append(subset)
    return cliques

# Brute force variant to check if all cliques are maximal
def vertex_maximal_cliques(cliques):
    # type: ([collections.OrderedDict[int, Vertex]]) -> [collections.OrderedDict[int, Vertex]]
    if len(cliques) != 0:
        i = 0
        while i < len(cliques):
            j = i+1
            while j < len(cliques):
                if cliques[i].keys() <= cliques[j].keys():
                    cliques.remove(cliques[i])
                    i -= 1
                    break
                elif cliques[i].keys() >= cliques[j].keys():
                    cliques.remove(cliques[j])
                    j -= 1
                j += 1
            i += 1
    return cliques


# Construct the graph from given file and use period/delta to find timeskips.
# We use this to split the instance
def get_graph_from_file_new(filename, delta, period, mode, del_small_clique, version):
    # type: (str, int, int, int, int) -> list[TempGraph]

    with open(filename) as f:
        g = TempGraph(filename)
        vertex_namespace = {}  # type: Dict[int, int]
        reverse_namespace = {}  # type: Dict[int, int]
        number_vertices = 0
        current_timestamp = None

        for line_num, line in enumerate(f):
            contents = line.strip().split()
            if len(contents) > 2:
                t, u, v = int(contents[0]), int(contents[1]), int(contents[2])

                if t != current_timestamp and current_timestamp is not None:
                    # not same timestamp
                    modulo = (t - current_timestamp) % period
                    if modulo != 0:
                        t += period - modulo
                current_timestamp = t

                if u == v:  # ignore self loops and isolated vertices
                    continue

                for x in [u, v]:
                    if x not in vertex_namespace:
                        vertex_namespace[x] = number_vertices
                        reverse_namespace[number_vertices] = x
                        number_vertices += 1
                        g.addVertex(vertex_namespace[x])

    for vertex in g.vertexList.values():
        vertex.name = reverse_namespace[vertex.id]

    for vertex in g.vertexList.values():
        vertex.name = vertex.id

    temp_graph = TempGraph(filename)
    temp_graph.vertexList = g.vertexList
    temp_graph.name_space = vertex_namespace
    temp_graph.reverse_name_space = reverse_namespace
    for vertex in temp_graph.vertexList.values():
        temp_graph.edgeList[vertex] = collections.OrderedDict()
    temp_graph.numVertices = g.numVertices
    temp_graph.mode = mode
    temp_graph.delta = delta
    temp_graph.del_small_clique = del_small_clique
    temp_graph.maximality_variant = version

    number_timestamps = 0
    temp_graph.set_t_min(0)


    with open(filename) as f:

        current_timestamp = None
        
        for line_num, line in enumerate(f):
            contents = line.strip().split()
            if not contents:
                continue
            t = int(contents[0])

            if current_timestamp is None: 
                temp_graph.time_space[0] = current_timestamp = t

            if len(contents) > 2:
                num_skipped = 0
                if t != current_timestamp:
                    modulo = (t - current_timestamp) % period
                    if modulo != 0:
                        t += period - modulo
                    num_skipped = int((t - current_timestamp)/period)

                if num_skipped > delta + 1:
                    # new graph split with this one
                    temp_graph.set_t_max(number_timestamps)
                    yield temp_graph.create_delta_graph(delta)

                    # start new graph:
                    temp_graph = TempGraph(filename)
                    temp_graph.set_t_min(number_timestamps+1)
                    temp_graph.vertexList = g.vertexList
                    temp_graph.name_space = vertex_namespace
                    temp_graph.reverse_name_space = reverse_namespace
                    for vertex in temp_graph.vertexList.values():
                        temp_graph.edgeList[vertex] = collections.OrderedDict()
                    temp_graph.numVertices = g.numVertices
                    temp_graph.mode = mode
                    temp_graph.delta = delta
                    temp_graph.del_small_clique = del_small_clique
                    temp_graph.maximality_variant = version


                for i in range(num_skipped):
                    current_timestamp += period
                    number_timestamps += 1
                    temp_graph.time_space[number_timestamps] = current_timestamp
                current_timestamp = t

                u, v = int(contents[1]), int(contents[2])
                if u != v:
                    temp_graph.addEdge(vertex_namespace[u], vertex_namespace[v], number_timestamps)

    temp_graph.set_t_max(number_timestamps)

    yield temp_graph.create_delta_graph(delta)


# Main method for reading graph
def read_temporal_graph(filename, delta, period=20, mode=0, del_small_clique=1, version=1):
    # type: (str, int, int, int, int, int) -> list[TempGraph]

    start = time.time()
    iter_g = get_graph_from_file_new(filename, delta, period, mode, del_small_clique, version)
    end = time.time()
    # print("Read input %.4f" % (end - start))
    return iter_g


def map_mode(mode):

    if mode == 0:
        return "alltime_max"
    if mode == 1:
        return "max_usually"
    if mode == 2:
        return "usally_average"
    if mode == 3:
        return "alltime_average"
    if mode == 4:
        return "average_alltime"
    if mode == 5:
        return "usally_max"
    return " "


def reverse_map_mode(mode):

    if mode == "alltime_max":
        return 0
    if mode == "max_usually":
        return 1
    if mode == "usually_average":
        return 2
    if mode == "alltime_average":
        return 3
    if mode == "average_alltime":
        return 4
    if mode == "usually_max":
        return 5

    return -1

def num_time_max_cliques(time_max_cliques):
    # type: (collections.OrderedDict[int, collections.OrderedDict[int, list[collections.OrderedDict[int, Vertex]]]]) -> None

    num = 0

    for a in time_max_cliques.keys():
        for b in time_max_cliques[a].keys():
            num += len(time_max_cliques[a][b])

    return num


def max_clique_size(time_max_cliques):
    max_size = 0
    for a in time_max_cliques.keys():
        for b in time_max_cliques[a].keys():
            for clique in time_max_cliques[a][b]:
                max_size = max(max_size, len(clique))
    return max_size



if __name__ == "__main__":
    sys.exit(main())
